/*- - - - - - - - - - - - - - - - - - - - - - - - - - - -
	BROWSE.CPP

	TBrowse Class

	Author:		Daniel BERMAN (CompuServe: 100567,3464)
	Date:			26 february 1995 (1.0)
	Revision:	08 march 1995 (1.01)
	Revision:	09 april 1995 (2.0)
	Revision:	28 may 1995 (2.01)
	Revision:	04 juin 1995 (2.02)
	Revision:	13 juin 1995 (2.03)
   Revision:	22 jully 1995 (2.1)

*/

#include <owl\owlpch.h>
#pragma hdrstop

#include "browse.h"


#define IDC_LISTBOX	24500
#define MULTIPLESEL	(LBS_MULTIPLESEL | LBS_EXTENDEDSEL)



//-- TColumn -------------------------------------------------------------------

struct TColumn
{
	int Id;
	string Text;

	TColumn() {}
	TColumn(int id, const string& text) {Id = id; Text = text;};

	bool operator == (const TColumn& col) const {return col.Id == Id;}
};
typedef TArrayAsVector<TColumn> TColumns;


//-- TBrowseList ---------------------------------------------------------------

class TBrowseList : public TListBox
{
public:

	TBrowseList(TWindow* parent, TModule* module);

protected:

	void EvHScroll(uint scrollCode, uint thumbPos, HWND hWndCtl);

private:

	DECLARE_RESPONSE_TABLE(TBrowseList);
};

DEFINE_RESPONSE_TABLE1(TBrowseList, TListBox)
	EV_WM_HSCROLL,
END_RESPONSE_TABLE;

TBrowseList::TBrowseList(TWindow* parent, TModule* module)
	: TListBox(parent, IDC_LISTBOX, 0, 0, 0, 0, module)
{
	Attr.Style &= ~(WS_BORDER | LBS_SORT);
	Attr.Style |= LBS_OWNERDRAWFIXED | LBS_NOINTEGRALHEIGHT;
}

void TBrowseList::EvHScroll(uint scrollCode, uint thumbPos, HWND hWndCtl)
{
	TListBox::EvHScroll(scrollCode, thumbPos, hWndCtl);
	TRect r = Parent->GetClientRect();
	r.bottom = r.top + TYPESAFE_DOWNCAST(Parent, TBrowse)->GetHeadersHeight();
	Parent->InvalidateRect(r);
}


//-- THeader -------------------------------------------------------------------

THeader::THeader()
{
	Id = Width = TextFormat = 0;
	Visible = false;
	Color = TColor(0, 0, 128);
	Rect.SetNull();
}

THeader::THeader(int id, const string& title, int width, bool visible)
{
	Id = id;
	Title = title;
	Width = width;
	Visible = visible;
	Color = TColor(0, 0, 128);
	TextFormat = DT_SINGLELINE | DT_VCENTER | DT_LEFT;
	Rect.SetNull();
}


//-- TBrowse -------------------------------------------------------------------

DEFINE_RESPONSE_TABLE1(TBrowse, TControl)
	EV_LBN_DBLCLK(IDC_LISTBOX, EvDblClk),
	EV_LBN_SELCHANGE(IDC_LISTBOX, EvSelChange),
	EV_WM_LBUTTONDOWN,
	EV_WM_LBUTTONUP,
	EV_WM_MOUSEMOVE,
	EV_WM_NCHITTEST,
	EV_WM_RBUTTONDOWN,
	EV_WM_RBUTTONUP,
	EV_WM_SETFOCUS,
	EV_WM_SIZE,
END_RESPONSE_TABLE;

TBrowse::TBrowse(TWindow* parent, int id, int x, int y, int w, int h, TModule* module)
	: TControl(parent, id, "", x, y, w, h, module)
{
	Attr.Style |= (WS_BORDER | WS_TABSTOP);
	Init();
}


TBrowse::TBrowse(TWindow* parent, int resourceId, TModule* module)
	: TControl(parent, resourceId, module)
{
	Init();
}

TBrowse::~TBrowse()
{
	delete NormalFont;
	delete BoldFont;
	delete Headers;
}

void TBrowse::AddColumn(int id, const char* title, int numChars, bool visible)
{
	if (numChars < 0)
		numChars = strlen(title);

	Headers->Add(new THeader(id, title, BoldCharWidth * numChars, visible));
}

int TBrowse::AddRow()
{
	int index;

	if ((index = List->AddString("")) != LB_ERR) {
		TColumns *columns = new TColumns(GetNumCols(), 0, 1);
		THeadersIterator i(*Headers);
		while (i) {
			columns->Add(TColumn(i.Current()->Id, ""));
			i++;
		}
		List->SetItemData(index, (uint32)columns);
	}

	return index;
}

void TBrowse::CalculateHeadersSize()
{
	TRect rect(0, 0, 0, HeadersHeight);
	int Extent = 0;

	THeadersIterator i(*Headers);
	while (i) {
		THeader* header = i++;
		if (header->Visible) {
			rect.right = rect.left + 4 + header->Width;
			header->Rect = rect;
			Extent += rect.Width() + 1;
			rect.left = rect.right + 1;
		}
	}

	if (Extent > List->GetHorizontalExtent())
		List->SetHorizontalExtent (HorzExtent=Extent);
	else if (Extent <= List->GetClientRect().Width()) {
		List->SetHorizontalExtent (HorzExtent=Extent);
		List->SetScrollPos(SB_HORZ, 0);
		List->ShowScrollBar(SB_HORZ, false);
	}
}

void TBrowse::ClearList()
{
	List->ClearList();
}

void TBrowse::DeleteItem(DELETEITEMSTRUCT far& deleteInfo)
{
	delete (TColumns*)deleteInfo.itemData;
}

int TBrowse::DeleteRow(int index)
{
	return List->DeleteString(index);
}

void TBrowse::DrawHeaderRect(TDC& dc, TRect& rect)
{
	dc.SaveDC();
	dc.SelectObject(TBrush(TColor(TColor::White)));
	dc.PatBlt(rect.left, rect.top, rect.Width() - 1, 1, PATCOPY);
	dc.PatBlt(rect.left, rect.top, 1, rect.Height() - 1, PATCOPY);
	dc.SelectObject(TBrush(TColor(TColor::Gray)));
	dc.PatBlt(rect.right - 1, rect.top, 1, rect.Height(), PATCOPY);
	dc.PatBlt(rect.left, rect.bottom - 1, rect.Width(), 1, PATCOPY);
	dc.RestoreDC();
}

void TBrowse::DrawItem(DRAWITEMSTRUCT far& drawInfo)
{
	if (drawInfo.CtlType == ODT_LISTBOX)
		switch (drawInfo.itemAction) {
			case ODA_DRAWENTIRE:
			case ODA_SELECT: {

				TDC dc(drawInfo.hDC);
				dc.SelectObject(*NormalFont);
				dc.SetBkColor(GetSysColor(drawInfo.itemState & ODS_SELECTED ? COLOR_HIGHLIGHT : COLOR_WINDOW));
				dc.SetTextColor(GetSysColor(drawInfo.itemState & ODS_SELECTED ? COLOR_HIGHLIGHTTEXT : COLOR_WINDOWTEXT));
				dc.TextRect(drawInfo.rcItem);

				TColumns* cols = (TColumns*)drawInfo.itemData;
				THeadersIterator i(*Headers);
				while (i) {
					THeader* header = i.Current();
					if (header->Visible) {
						TRect rect = header->Rect.InflatedBy(-1, -1);
						rect.top = drawInfo.rcItem.top;
						rect.bottom = drawInfo.rcItem.bottom;
						int loc = cols->Find(TColumn(header->Id, ""));
						if (loc != INT_MAX)
							dc.DrawText((*cols)[loc].Text.c_str(), -1, rect, DT_SINGLELINE | DT_VCENTER);
					}
					i++;
				}

				if (drawInfo.itemState & ODS_FOCUS)
					dc.DrawFocusRect(drawInfo.rcItem);

				break;

			}

			case ODA_FOCUS: {
				TDC dc(drawInfo.hDC);
				dc.DrawFocusRect(drawInfo.rcItem);
			}
		}
}

void TBrowse::DrawSizingLine(TPoint& point)
{
	TRect rect = List->GetClientRect();
	TWindowDC dc(*List);

	TPen pen(TColor::Black, 1, PS_DOT);
	dc.SetROP2(R2_XORPEN);
	dc.SelectObject(pen);

	dc.MoveTo(PrevSizingRect.TopLeft());
	dc.LineTo(PrevSizingRect.BottomRight());

	dc.MoveTo(point.x, rect.top);
	dc.LineTo(point.x, rect.bottom);

	PrevSizingRect = TRect(point.x, rect.top, point.x, rect.bottom);
}

void TBrowse::EvDblClk()
{
	if (Parent)
		Parent->SendMessage(WM_COMMAND, GetId(), MAKELPARAM(HWindow, LBN_DBLCLK));
}

void TBrowse::EvLButtonDown(uint modKeys, TPoint& point)
{
	Capture = GetHeaderFromPoint(point);

	if (Capture != NULL) {
		SetCapture();
		ReSizing = true;
		Dragging = false;
		StartDrawingSizingLine(point);
	}
	TControl::EvLButtonDown(modKeys, point);
}

void TBrowse::EvLButtonUp(uint modKeys, TPoint& point)
{
	if (Capture) {
		if (ReSizing) {
			StopDrawingSizingLine(point);
			TRect rect = Capture->Rect.OffsetBy(-ViewportOrgX, 0);
			if (point.x < rect.right - 1 || point.x > rect.right + 1) {
				if (point.x < rect.left)
					point.x = rect.left;
				Capture->Width = point.x - rect.left - 4;
				CalculateHeadersSize();
				Invalidate();
			}
		}
		Capture = NULL;
		ReSizing = Dragging = false;
		SetCursor(0, IDC_ARROW);
		ReleaseCapture();
	}
	TControl::EvLButtonUp(modKeys, point);
}

void TBrowse::EvMouseMove(uint modKeys, TPoint& point)
{
	if (!Capture)
		if (GetHeaderFromPoint(point) != NULL)
			SetCursor(0, IDC_SIZEWE);
		else
			SetCursor(0, IDC_ARROW);
	else
		if (ReSizing)
			DrawSizingLine(point);

	TControl::EvMouseMove(modKeys, point);
}

uint TBrowse::EvNCHitTest(TPoint& /*point*/)
{
	return HTCLIENT;
}

void TBrowse::EvRButtonDown(uint modKeys, TPoint& point)
{
	if (Parent) {
		MapWindowPoints(*Parent, &point, 1);
		Parent->SendMessage(WM_RBUTTONDOWN, modKeys, MAKELPARAM(point.x, point.y));
	}
}

void TBrowse::EvRButtonUp(uint modKeys, TPoint& point)
{
	if (Parent) {
		MapWindowPoints(*Parent, &point, 1);
		Parent->SendMessage(WM_RBUTTONUP, modKeys, MAKELPARAM(point.x, point.y));
	}
}

void TBrowse::EvSelChange()
{
	if (Parent)
		Parent->SendMessage(WM_COMMAND, GetId(), MAKELPARAM(HWindow, LBN_SELCHANGE));
}

void TBrowse::EvSetFocus(HWND hWndLostFocus)
{
	TControl::EvSetFocus(hWndLostFocus);
	List->SetFocus();
}

void TBrowse::EvSize(uint sizeType, TSize& size)
{
	TControl::EvSize(sizeType, size);

	TRect rect = GetClientRect();
	rect.top += HeadersHeight;
	List->MoveWindow(rect);

	Invalidate();
}

bool TBrowse::GetColumnText(int id, string& text, int index)
{
	bool bOk = false;

	uint32 itemData = List->GetItemData(index);
	if ((long)itemData != LB_ERR) {
		TColumns* cols = (TColumns*)itemData;
		int loc = cols->Find(TColumn(id, ""));
		if (loc != INT_MAX) {
			text = (*cols)[loc].Text;
			bOk = true;
		}
	}

	return bOk;
}

int TBrowse::GetCount() const
{
	return List->GetCount();
}

bool TBrowse::GetHeaderColor(int id, TColor& color)
{
	int loc = Headers->Find(&THeader(id, "", 0, true));
	if (loc != INT_MAX) {
		color = (*Headers)[loc]->Color;
		return true;
	}
	return false;
}

THeader* TBrowse::GetHeaderFromPoint(TPoint& point)
{
	THeadersIterator i(*Headers);
	while (i) {
		THeader* header = i.Current();
		if (header->Visible) {
			TRect rect = header->Rect.OffsetBy(-ViewportOrgX, 0);
			if (point.x >= rect.right - 1 && point.x <= rect.right + 1	&&
				 point.y >= rect.top  && point.y <= rect.bottom)
				return header;
		}
		i++;
	}
	return NULL;
}

int TBrowse::GetNumCols()
{
	return Headers->GetItemsInContainer();
}

bool TBrowse::GetSel(int index) const
{
	return List->GetSel(index);
}

int TBrowse::GetSelCount() const
{
	return List->GetSelCount();
}

int TBrowse::GetSelIndex() const
{
	return List->GetSelIndex();
}

int TBrowse::GetSelIndexes(int* indexes, int maxCount) const
{
	return List->GetSelIndexes(indexes, maxCount);
}

int TBrowse::GetTopIndex() const
{
	return List->GetTopIndex();
}

void TBrowse::Init()
{
	ReSizing = Dragging = false;
	Capture = NULL;

	NormalFont = new TFont("MS Sans Serif", -11);

	LOGFONT lf;
	NormalFont->GetObject(lf);
	lf.lfWeight = FW_BOLD;
	BoldFont = new TFont(&lf);

	{
		TRect rc;
		TScreenDC dc;
		dc.SelectObject(*BoldFont);
		dc.DrawText("Abc", -1, rc, DT_CENTER | DT_CALCRECT);
		HeadersHeight = rc.Height() + 4;

		TEXTMETRIC tm;
		dc.GetTextMetrics(tm);
		BoldCharWidth = (tm.tmAveCharWidth + tm.tmMaxCharWidth) / 2;
	}

	Headers = new THeaders(2, 0, 1);

	SetBkgndColor(TColor::LtGray);

	EnableTransfer();
}

bool TBrowse::IsColumnVisible(int id)
{
	int loc = Headers->Find(&THeader(id, "", 0, true));
	if (loc != INT_MAX) {
		return (*Headers)[loc]->Visible;
	}
	return false;
}

void TBrowse::MeasureItem(MEASUREITEMSTRUCT far& measureInfo)
{
	if (measureInfo.CtlType == ODT_LISTBOX) {
		TScreenDC dc;
		TEXTMETRIC tm;
		dc.GetTextMetrics(tm);
		measureInfo.itemHeight = tm.tmHeight + tm.tmExternalLeading;
	}
}

bool TBrowse::ModifyColumnText(int id, const char* text, int index)
{
	if (SetColumnText(id, text, index)) {
		List->Invalidate(false);
		return true;
	}
	return false;
}

void TBrowse::Paint(TDC& dc, bool /*erase*/, TRect& rect)
{
	CalculateHeadersSize();

	ViewportOrgX = (float(HorzExtent) - List->GetClientRect().Width()) / 100
						* List->GetScrollPos(SB_HORZ);
	TPoint offset(-ViewportOrgX, 0);
	dc.SetViewportOrg(offset);
	rect -= offset;

	dc.SelectObject(*BoldFont);
	dc.SetBkMode(TRANSPARENT);

	TRect thisRect(GetClientRect());
	thisRect.bottom = thisRect.top + HeadersHeight;
	dc.SelectClipRgn(thisRect);

	THeadersIterator i(*Headers);
	while (i) {
		THeader* header = i.Current();
		if (header->Visible) {
			dc.SetTextColor(header->Color);
			DrawHeaderRect(dc, header->Rect);
			TRect r = header->Rect.InflatedBy(-2, -2);
			dc.DrawText(header->Title.c_str(), -1, r, header->TextFormat);
		}
		i++;
	}
}

bool TBrowse::SetColumnText(int id, const char* text, int index)
{
	bool bOk = false;

	if (GetCount() < 1)
		return bOk;

	if (index == -1)
		index = GetCount() - 1;

	uint32 itemData = List->GetItemData(index);
	if ((long)itemData != LB_ERR) {
		TColumns *cols = (TColumns*)itemData;
		int loc = cols->Find(TColumn(id, ""));
		if (loc != INT_MAX) {
			(*cols)[loc].Text = text;
			bOk = true;
		}
	}

	return bOk;
}

bool TBrowse::SetColumnWidth(int id, int numChars)
{
	int loc = Headers->Find(&THeader(id, "", 0, true));
	if (loc != INT_MAX) {
		(*Headers)[loc]->Width = BoldCharWidth * numChars;
		Invalidate();
		return true;
	}
	return false;
}

bool TBrowse::SetHeaderColor(int id, TColor color)
{
	int loc = Headers->Find(&THeader(id, "", 0, true));
	if (loc != INT_MAX) {
		(*Headers)[loc]->Color = color;
		Invalidate();
		return true;
	}
	return false;
}

bool TBrowse::SetHeaderTextFormat(int id, uint16 textFormat)
{
	int loc = Headers->Find(&THeader(id, "", 0, true));
	if (loc != INT_MAX) {
		(*Headers)[loc]->TextFormat = textFormat;
		Invalidate();
		return true;
	}
	return false;
}

void TBrowse::SetHeadersHeight(int height)
{
	HeadersHeight = height;
	Invalidate();
}

int TBrowse::SetSel(int index, bool select)
{
	return List->SetSel(index, select);
}

int TBrowse::SetSelIndex(int index)
{
	return List->SetSelIndex(index);
}

int TBrowse::SetTopIndex(int index)
{
	return List->SetTopIndex(index);
}

void TBrowse::SetupWindow()
{
	List = new TBrowseList(this, GetModule());

	if (Attr.Style & TBS_MULTIPLESEL)
		List->Attr.Style |= LBS_MULTIPLESEL;
	if (Attr.Style & TBS_EXTENDEDSEL)
		List->Attr.Style |= LBS_EXTENDEDSEL;

	TControl::SetupWindow();

	ClearFlag(wfPredefinedClass);

	TRect rect = GetClientRect();
	rect.top += HeadersHeight;
	List->MoveWindow(rect);
}

bool TBrowse::ShowColumn(int id, int visible)
{
	int loc = Headers->Find(&THeader(id, "", 0, true));
	if (loc != INT_MAX) {
		(*Headers)[loc]->Visible = visible;
		Invalidate();
		return true;
	}
	return false;
}

void TBrowse::StartDrawingSizingLine(TPoint& point)
{
	TRect rect = List->GetClientRect();
	TWindowDC dc(*List);

	TPen pen(TColor::Black, 1, PS_DOT);
	dc.SetROP2(R2_XORPEN);
	dc.SelectObject(pen);

	dc.MoveTo(point.x, rect.top);
	dc.LineTo(point.x, rect.bottom);

	PrevSizingRect = TRect(point.x, rect.top, point.x, rect.bottom);
}

void TBrowse::StopDrawingSizingLine(TPoint& /*point*/)
{
	TRect rect = List->GetClientRect();
	TWindowDC dc(*List);

	TPen pen(TColor::Black, 1, PS_DOT);
	dc.SetROP2(R2_XORPEN);
	dc.SelectObject(pen);

	dc.MoveTo(PrevSizingRect.TopLeft());
	dc.LineTo(PrevSizingRect.BottomRight());
}

uint TBrowse::Transfer(void *buffer, TTransferDirection direction)
{
	long style = List->GetWindowLong(GWL_STYLE);
	TBrowseData* browseData = (TBrowseData*)buffer;

	if (direction == tdGetData) {
		browseData->Clear();
		for (int i = 0; i < GetCount(); i++) {
			TColumns* src = (TColumns*)List->GetItemData(i);
			browseData->AddRow();
			for (int j = 0; j < src->GetItemsInContainer(); j++)
				browseData->SetColumnText((*src)[j].Id, (*src)[j].Text.c_str());
		}
		browseData->ResetSelections();
		if (!(style & MULTIPLESEL)) {
			browseData->Select(GetSelIndex());
		}
		else {
			int selCount = GetSelCount();
			if (selCount > 0) {
				int* selections = new int[selCount];
				GetSelIndexes(selections, selCount);
				for (int selIndex = 0; selIndex < selCount; selIndex++)
					browseData->Select(selections[selIndex]);
				delete selections;
			}
		}
	}
	else if (direction == tdSetData) {
		ClearList();
		const int noSelection = -1;
		int  selCount = browseData->GetSelCount();
		int  selIndex = noSelection;
		for (int i = 0; i < browseData->GetItemDatas().GetItemsInContainer(); i++) {
			int index = AddRow();
			TColumns* cols = (TColumns*)browseData->GetItemDatas()[i];
			for (int j = 0; j < cols->GetItemsInContainer(); j++)
				SetColumnText((*cols)[j].Id, (*cols)[j].Text.c_str(), index);
			if (style & MULTIPLESEL) {
				for (int k = 0; k < selCount; k++)
					if (browseData->GetSelIndices()[k] == i) {
						SetSel(index, true);
						break;
					}
			}
			else {
				if (browseData->GetSelIndices()[0] == i)
					selIndex = index;
				else
					if (selIndex != noSelection && index <= selIndex)
						selIndex++;
			}
		}
		if (selIndex != noSelection && !(style & MULTIPLESEL))
			SetSelIndex(selIndex);
	}

	return sizeof(TBrowseData);
}


//-- TBrowseData ---------------------------------------------------------------

TBrowseData::TBrowseData()
	: ItemDatas(10, 0, 10), SelIndices(10, 0, 10)
{
	NumCols = 0;
}

TBrowseData::TBrowseData(int numCols)
	: ItemDatas(10, 0, 10), SelIndices(10, 0, 10)
{
	NumCols = numCols;
}

void TBrowseData::AddRow(bool isSelected)
{
	TColumns *columns = new TColumns(NumCols, 0, 1);
	ItemDatas.Add((uint32*)columns);
	if (isSelected)
		Select(ItemDatas.GetItemsInContainer()-1);
}

void TBrowseData::SetColumnText(int id, const char* text)
{
	TColumns *cols = (TColumns*)ItemDatas[ItemDatas.GetItemsInContainer()-1];
	int loc = cols->Find(TColumn(id, ""));
	if (loc != INT_MAX)
		(*cols)[loc].Text = text;
	else
		cols->Add(TColumn(id, text));
}

void TBrowseData::Select(int index)
{
	if (index != LB_ERR)
		SelIndices.Add(index);
}


//-- TRowArray -----------------------------------------------------------------

TRowArray::TRowArray(int upper, int lower, int delta)
	: Data(upper, lower, delta)
{
}

int TRowArray::Add(uint32* t)
{
	return Data.Add(t);
}

unsigned TRowArray::ArraySize() const
{
	return Data.ArraySize();
}

int TRowArray::Detach(uint32* t)
{
	return Data.Detach(t);
}

int TRowArray::Detach(int loc)
{
	return Data.Detach(loc);
}

int TRowArray::Destroy(uint32* t)
{
	return Detach(t);
}

int TRowArray::Destroy(int loc)
{
	return Detach(loc);
}

int TRowArray::Find(uint32* t) const
{
	return Data.Find(t);
}

uint32* TRowArray::FirstThat(CondFunc cond, void* args) const
{
	return Data.FirstThat(cond, args);
}

void TRowArray::Flush()
{
	Data.Flush();
}

void TRowArray::ForEach(IterFunc iter, void* args)
{
	Data.ForEach(iter, args);
}

unsigned TRowArray::GetItemsInContainer() const
{
	return Data.GetItemsInContainer();
}

int TRowArray::HasMember(uint32* t) const
{
	return Data.HasMember(t);
}

int TRowArray::IsEmpty() const
{
	return Data.IsEmpty();
}

int TRowArray::IsFull() const
{
	return Data.IsFull();
}

uint32* TRowArray::LastThat(CondFunc cond, void* args) const
{
	return Data.LastThat(cond, args);
}

int TRowArray::LowerBound() const
{
	return Data.LowerBound();
}

int TRowArray::UpperBound() const
{
	return Data.UpperBound();
}

uint32* TRowArray::operator [](int loc)
{
	return Data[loc];
}

uint32* TRowArray::operator [](int loc) const
{
	return Data[loc];
}


//-- TRowArrayIterator----------------------------------------------------------

TRowArrayIterator::TRowArrayIterator(const TRowArray& array)
	: TIArrayAsVectorIterator<uint32>(array.Data)
{
}

bool TRowArrayIterator::GetColumnText(int id, string& text)
{
	TColumns* cols = (TColumns*)Current();
	int loc = cols->Find(TColumn(id, ""));
	if (loc != INT_MAX) {
		text = (*cols)[loc].Text;
		return true;
	}
	return false;
}


