/*- - - - - - - - - - - - - - - - - - - - - - - - - - - -
	BROWSE.H

	TBrowse Class

	Author:		Daniel BERMAN (CompuServe: 100567,3464)
	Date:			26 february 1995 (1.0)
	Revision:	08 march 1995 (1.01)
	Revision:	09 april 1995 (2.0)
	Revision:	28 may 1995 (2.01)
	Revision:	04 juin 1995 (2.02)
	Revision:	13 juin 1995 (2.03)
	Revision:	22 jully 1995 (2.1)

*/

#ifndef _TBrowse_h
#define _TBrowse_h

// TBrowse styles
#define TBS_MULTIPLESEL		0x0040L
#define TBS_EXTENDEDSEL		0x0080L

#ifndef RC_INVOKED

#include <owl\owlpch.h>
#pragma hdrstop

#include <classlib\arrays.h>



struct THeader
{
	int Id;
	string Title;
	bool Visible;
	int Width;
	TRect Rect;
	uint16 TextFormat;
	TColor Color;

	THeader();
	THeader(int id, const string& title, int width, bool visible);

	bool operator == (const THeader& header) const {return header.Id == Id;}
};
typedef TIArrayAsVector<THeader> THeaders;
typedef TIArrayAsVectorIterator<THeader> THeadersIterator;


class TBrowseList;


class TBrowse : public TControl
{
public:

	TBrowse(TWindow* parent, int id, int x, int y, int w, int h, TModule* module = 0);
	TBrowse(TWindow* parent, int resourceId, TModule* module = 0);
	virtual ~TBrowse();

	virtual void	AddColumn(int id, const char* title, int numChars = -1, bool visible = true);
	virtual int		AddRow();
	virtual void	ClearList();
	virtual int		DeleteRow(int index);
	bool				GetColumnText(int id, string& text, int index);
	virtual int		GetCount() const;
	bool				GetHeaderColor(int id, TColor& color);
	int				GetHeadersHeight() {return HeadersHeight;}
	int				GetNumCols();
	bool				GetSel(int index) const;
	int				GetSelCount() const;
	int				GetSelIndex() const;
	int				GetSelIndexes(int* indexes, int maxCount) const;
	int				GetTopIndex() const;
	bool				IsColumnVisible(int id);
	virtual bool	ModifyColumnText(int id, const char* text, int index);
	virtual bool	SetColumnText(int id, const char* text, int index = -1);
	bool				SetColumnWidth(int id, int numChars);
	bool				SetHeaderColor(int id, TColor color);
	bool				SetHeaderTextFormat(int id, uint16 uiTextFormat);
	void				SetHeadersHeight(int height);
	int				SetSel(int index, bool select);
	virtual int		SetSelIndex(int index);
	int				SetTopIndex(int index);
	bool				ShowColumn(int id, int visible = true);
	virtual uint	Transfer(void *buffer, TTransferDirection direction);

protected:

	virtual void		CalculateHeadersSize();
	virtual void		DeleteItem(DELETEITEMSTRUCT far& deleteInfo);
	virtual void		DrawHeaderRect(TDC& dc, TRect& rect);
	virtual void		DrawItem(DRAWITEMSTRUCT far& drawInfo);
	virtual void		DrawSizingLine(TPoint& point);
	void					EvDblClk();
	void					EvLButtonDown(uint modKeys, TPoint& point);
	void					EvLButtonUp(uint modKeys, TPoint& point);
	void					EvMouseMove(uint modKeys, TPoint& point);
   uint					EvNCHitTest(TPoint& point);
	void					EvRButtonDown(uint modKeys, TPoint& point);
	void					EvRButtonUp(uint modKeys, TPoint& point);
	void					EvSelChange();
	void					EvSetFocus(HWND hWndLostFocus);
	void					EvSize(uint sizeType, TSize& size);
	virtual THeader*	GetHeaderFromPoint(TPoint& point);
	virtual void		MeasureItem(MEASUREITEMSTRUCT far& measureInfo);
	virtual void		Paint(TDC& dc, bool erase, TRect& rect);
	virtual void		SetupWindow();
	virtual void		StartDrawingSizingLine(TPoint& point);
	virtual void		StopDrawingSizingLine(TPoint& point);

private:

	TFont* NormalFont;
	TFont* BoldFont;
	int HeadersHeight;
	int BoldCharWidth;
	int HorzExtent;
	int ViewportOrgX;

	THeader* Capture;
	bool ReSizing;
	bool Dragging;
	TRect PrevSizingRect;

	THeaders* Headers;
	TBrowseList* List;

	void Init();

	DECLARE_RESPONSE_TABLE(TBrowse);
};


class TRowArray
{
public:

	typedef void	(*IterFunc)(uint32&, void*);
	typedef int		(*CondFunc)(const uint32&, void*);

	TRowArray(int upper, int lower, int delta);

	int 		Add(uint32* t);
	unsigned ArraySize() const;
	int		Detach(uint32* t);
	int		Detach(int loc);
	int		Destroy(uint32* t);
	int		Destroy(int loc);
	int		Find(uint32* t) const;
	uint32*	FirstThat(CondFunc cond, void* args) const;
	void		Flush();
	void		ForEach(IterFunc iter, void* args);
	unsigned	GetItemsInContainer() const;
	int		HasMember(uint32* t) const;
	int 		IsEmpty() const;
	int 		IsFull() const;
	uint32*	LastThat(CondFunc cond, void* args) const;
	int 		LowerBound() const;
	int 		UpperBound() const;

	uint32*	operator [](int loc);
	uint32*	operator [](int loc) const;

private:

	TIArrayAsVector<uint32> Data;

	friend class TRowArrayIterator;
};


class TRowArrayIterator : public TIArrayAsVectorIterator<uint32>
{
public:

	TRowArrayIterator(const TRowArray& array);

	bool GetColumnText(int id, string& text);
};


class TBrowseData
{
public:

	TBrowseData();
	TBrowseData(int numCols);

	void				AddRow(bool isSelected = false);
	void				Clear() {ItemDatas.Flush(); ResetSelections();}
	int				GetNumRows() {return ItemDatas.GetItemsInContainer();}
	TRowArray&		GetItemDatas() {return ItemDatas;}
	int				GetNumCols(int numCols) {return NumCols;}
	int				GetSelCount() const {return SelIndices.GetItemsInContainer();}
	TIntArray&		GetSelIndices() {return SelIndices;}
	void				ResetSelections() {SelIndices.Flush();}
	void				Select(int index);
	void				SetColumnText(int id, const char* text);
	void				SetNumCols(int numCols) {NumCols = numCols;}


protected:

	int NumCols;
	TRowArray ItemDatas;
	TIntArray SelIndices;
};


#endif	// RC_INVOKED

#endif	// _TBrowse_h
